@@ -5,7 +5,7 @@
 ;; Author: Omar Antolín Camarena <omar@matem.unam.mx>, Daniel Mendler <mail@daniel-mendler.de>
 ;; Maintainer: Omar Antolín Camarena <omar@matem.unam.mx>, Daniel Mendler <mail@daniel-mendler.de>
 ;; Created: 2020
-;; Version: 0.6
+;; Version: 0.9
 ;; Package-Requires: ((emacs "26.1"))
 ;; Homepage: https://github.com/minad/marginalia
 
@@ -39,16 +39,16 @@
 (defgroup marginalia nil
   "Enrich existing commands with completion annotations."
   :group 'convenience
+  :group 'minibuffer
   :prefix "marginalia-")
 
 (defcustom marginalia-truncate-width 80
   "Maximum truncation width of annotation fields.
 
-This value is adjusted in the `minibuffer-setup-hook' depending
-on the `window-width'."
+This value is adjusted depending on the `window-width'."
   :type 'integer)
 
-(defcustom marginalia-separator-threshold 120
+(defcustom marginalia-separator-threshold 160
   "Use wider separator for window widths larger than this value."
   :type 'integer)
 
@@ -69,12 +69,11 @@ It can also be set to an integer value of 1 or larger to force an offset."
   "Use whitespace margin for window widths larger than this value."
   :type 'integer)
 
-(defvar marginalia-annotators nil)
-(defvar marginalia-annotators-light nil)
-(defvar marginalia-annotators-heavy nil)
-(make-obsolete-variable 'marginalia-annotators "Deprecated in favor of `marginalia-annotator-registry'." "0.5")
-(make-obsolete-variable 'marginalia-annotators-light "Deprecated in favor of `marginalia-annotator-registry'." "0.5")
-(make-obsolete-variable 'marginalia-annotators-heavy "Deprecated in favor of `marginalia-annotator-registry'." "0.5")
+(defcustom marginalia-max-relative-age (* 60 60 24 14)
+  "Maximum relative age in seconds displayed by the file annotator.
+
+Set to `most-positive-fixnum' to always use a relative age, or 0 to never show a relative age."
+  :type 'integer)
 
 (defcustom marginalia-annotator-registry
   (mapcar
@@ -83,12 +82,12 @@ It can also be set to an integer value of 1 or larger to force an offset."
      (embark-keybinding marginalia-annotate-embark-keybinding)
      (customize-group marginalia-annotate-customize-group)
      (variable marginalia-annotate-variable)
+     (function marginalia-annotate-function)
      (face marginalia-annotate-face)
      (color marginalia-annotate-color)
      (unicode-name marginalia-annotate-char)
      (minor-mode marginalia-annotate-minor-mode)
      (symbol marginalia-annotate-symbol)
-     (variable marginalia-annotate-variable)
      (environment-variable marginalia-annotate-environment-variable)
      (input-method marginalia-annotate-input-method)
      (coding-system marginalia-annotate-coding-system)
@@ -122,17 +121,24 @@ determine it."
     ("\\<M-x\\>" . command)
     ("\\<package\\>" . package)
     ("\\<bookmark\\>" . bookmark)
-    ("\\<face\\>" . face)
     ("\\<color\\>" . color)
+    ("\\<face\\>" . face)
     ("\\<environment variable\\>" . environment-variable)
+    ("\\<function\\>" . function)
     ("\\<variable\\>" . variable)
     ("\\<input method\\>" . input-method)
     ("\\<charset\\>" . charset)
     ("\\<coding system\\>" . coding-system)
-    ("\\<minor mode\\>" . minor-mode))
+    ("\\<minor mode\\>" . minor-mode)
+    ("\\<[Ll]ibrary\\>" . library))
   "Associates regexps to match against minibuffer prompts with categories."
   :type '(alist :key-type regexp :value-type symbol))
 
+(defcustom marginalia-censor-variables
+  '("pass")
+  "The values of variables matching any of these regular expressions is not shown."
+  :type '(repeat (choice symbol regexp)))
+
 (defcustom marginalia-command-categories
   '((imenu . imenu))
   "Associate commands with a completion category."
@@ -141,7 +147,7 @@ determine it."
 (defcustom marginalia-bookmark-type-transformers
   `(("\\`bookmark-\\(.*?\\)-handler\\'" . "\\1")
     ("default" . "File")
-    ("\\`\\(.*?\\)-bookmark-jump\\(?:-handler\\)?\\'" . "\\1")
+    ("\\`\\(.*?\\)-+bookmark-jump\\(?:-handler\\)?\\'" . "\\1")
     (".*" . ,#'capitalize))
   "List of bookmark type transformers."
   :type '(alist :key-type regexp :value-type (choice string function)))
@@ -153,19 +159,19 @@ determine it."
 
 (defface marginalia-key
   '((t :inherit font-lock-keyword-face))
-  "Face used to highlight keys in `marginalia-mode'.")
+  "Face used to highlight keys.")
 
 (defface marginalia-type
   '((t :inherit marginalia-key))
-  "Face used to highlight types in `marginalia-mode'.")
+  "Face used to highlight types.")
 
 (defface marginalia-char
   '((t :inherit marginalia-key))
-  "Face used to highlight char in `marginalia-mode'.")
+  "Face used to highlight character annotations.")
 
 (defface marginalia-lighter
   '((t :inherit marginalia-size))
-  "Face used to highlight lighters in `marginalia-mode'.")
+  "Face used to highlight minor mode lighters.")
 
 (defface marginalia-on
   '((t :inherit success))
@@ -177,60 +183,114 @@ determine it."
 
 (defface marginalia-documentation
   '((t :inherit completions-annotations))
-  "Face used to highlight documentation string in `marginalia-mode'.")
+  "Face used to highlight documentation strings.")
 
-(defface marginalia-variable
+(defface marginalia-value
   '((t :inherit marginalia-key))
-  "Face used to highlight variable values in `marginalia-mode'.")
+  "Face used to highlight general variable values.")
+
+(defface marginalia-null
+  '((t :inherit font-lock-comment-face))
+  "Face used to highlight null or unbound variable values.")
+
+(defface marginalia-true
+  '((t :inherit font-lock-builtin-face))
+  "Face used to highlight true variable values.")
+
+(defface marginalia-function
+  '((t :inherit font-lock-function-name-face))
+  "Face used to highlight function symbols.")
+
+(defface marginalia-symbol
+  '((t :inherit font-lock-type-face))
+  "Face used to highlight general symbols.")
+
+(defface marginalia-list
+  '((t :inherit font-lock-constant-face))
+  "Face used to highlight list expressions.")
 
 (defface marginalia-mode
   '((t :inherit marginalia-key))
-  "Face used to highlight major modes in `marginalia-mode'.")
+  "Face used to highlight buffer major modes.")
 
 (defface marginalia-date
   '((t :inherit marginalia-key))
-  "Face used to highlight dates in `marginalia-mode'.")
+  "Face used to highlight dates.")
 
 (defface marginalia-version
   '((t :inherit marginalia-number))
-  "Face used to highlight package version in `marginalia-mode'.")
+  "Face used to highlight package versions.")
 
 (defface marginalia-archive
   '((t :inherit warning))
-  "Face used to highlight package archives in `marginalia-mode'.")
+  "Face used to highlight package archives.")
 
 (defface marginalia-installed
   '((t :inherit success))
-  "Face used to highlight package status in `marginalia-mode'.")
+  "Face used to highlight the status of packages.")
 
 (defface marginalia-size
   '((t :inherit marginalia-number))
-  "Face used to highlight sizes in `marginalia-mode'.")
+  "Face used to highlight sizes.")
 
 (defface marginalia-number
   '((t :inherit font-lock-constant-face))
-  "Face used to highlight char in `marginalia-mode'.")
+  "Face used to highlight numeric values.")
+
+(defface marginalia-string
+  '((t :inherit font-lock-string-face))
+  "Face used to highlight string values.")
 
 (defface marginalia-modified
   '((t :inherit font-lock-negation-char-face))
-  "Face used to highlight modification indicators in `marginalia-mode'.")
+  "Face used to highlight buffer modification indicators.")
 
 (defface marginalia-file-name
   '((t :inherit marginalia-documentation))
-  "Face used to highlight file names in `marginalia-mode'.")
-
-(defface marginalia-file-modes
-  '((t :inherit font-lock-string-face))
-  "Face used to highlight file modes in `marginalia-mode'.")
+  "Face used to highlight file names.")
 
 (defface marginalia-file-owner
   '((t :inherit font-lock-preprocessor-face))
-  "Face used to highlight file owners in `marginalia-mode'.")
+  "Face used to highlight file owner and group names.")
+
+(defface marginalia-file-priv-no
+  '((t :inherit shadow))
+  "Face used to highlight the no file privilege attribute.")
+
+(defface marginalia-file-priv-dir
+  '((t :inherit font-lock-keyword-face))
+  "Face used to highlight the dir file privilege attribute.")
+
+(defface marginalia-file-priv-link
+  '((t :inherit font-lock-keyword-face))
+  "Face used to highlight the link file privilege attribute.")
+
+(defface marginalia-file-priv-read
+  '((t :inherit font-lock-type-face))
+  "Face used to highlight the read file privilege attribute.")
+
+(defface marginalia-file-priv-write
+  '((t :inherit font-lock-builtin-face))
+  "Face used to highlight the write file privilege attribute.")
+
+(defface marginalia-file-priv-exec
+  '((t :inherit font-lock-function-name-face))
+  "Face used to highlight the exec file privilege attribute.")
+
+(defface marginalia-file-priv-other
+  '((t :inherit font-lock-constant-face))
+  "Face used to highlight some other file privilege attribute.")
+
+(defface marginalia-file-priv-rare
+  '((t :inherit font-lock-variable-name-face))
+  "Face used to highlight a rare file privilege attribute.")
 
 ;;;; Pre-declarations for external packages
 
 (defvar bookmark-alist)
-(declare-function bookmark-get-bookmark-record "bookmark")
+(declare-function bookmark-get-handler "bookmark")
+(declare-function bookmark-get-filename "bookmark")
+(declare-function bookmark-get-front-context-string "bookmark")
 
 (defvar package--builtins)
 (defvar package-archive-contents)
@@ -250,13 +310,24 @@ determine it."
 
 ;;;; Marginalia mode
 
+(defvar marginalia--fontified-file-modes nil
+  "List of fontified file modes.")
+
+(defvar-local marginalia--cache nil
+  "The cache, pair of list and hashtable.")
+
+(defvar marginalia--cache-size 100
+  "Size of the cache, set to 0 to disable the cache.
+Disabling the cache is useful on non-incremental UIs like default completion or
+for performance profiling of the annotators.")
+
 (defvar marginalia--separator "    "
   "Field separator.")
 
-(defvar marginalia--margin nil
+(defvar marginalia--margin 0
   "Right margin.")
 
-(defvar-local marginalia--this-command nil
+(defvar-local marginalia--command nil
   "Last command symbol saved in order to allow annotations.")
 
 (defvar-local marginalia--base-position 0
@@ -271,18 +342,16 @@ determine it."
    (if-let (pos (string-match-p "\n" str))
        (substring str 0 pos)
      str)
-   width 0 32 "…"))
+   width 0 32 t))
 
 (defun marginalia--align (str)
   "Align STR at the right margin."
   (unless (string-blank-p str)
-    (when marginalia--margin
-      (setq str (concat str marginalia--margin)))
     (concat " "
             (propertize
              " "
              'display
-             `(space :align-to (- right ,marginalia-align-offset ,(string-width str))))
+             `(space :align-to (- right ,marginalia--margin ,(string-width str))))
             str)))
 
 (cl-defmacro marginalia--field (field &key truncate format face width)
@@ -365,16 +434,19 @@ WIDTH is the format width. This can be specified as alternative to FORMAT."
 Function:
 f function
 c command
+C interactive-only command
 m macro
+p pure
+s side-effect-free
+@ autoloaded
 ! advised
-o obsolete
+- obsolete
 
 Variable:
-u custom
+u custom (U modified compared to global value)
 v variable
-l local
-* modified
-o obsolete
+l local (L modified compared to default value)
+- obsolete
 
 Other:
 a face
@@ -384,18 +456,31 @@ t cl-type"
    (concat
     (when (fboundp s)
       (concat
-       (and (get s 'byte-obsolete-info) "o")
        (cond
-        ((commandp s) "c")
+        ((get s 'pure) "p")
+        ((get s 'side-effect-free) "s"))
+       (cond
+        ((commandp s) (if (get s 'interactive-only) "C" "c"))
         ((eq (car-safe (symbol-function s)) 'macro) "m")
         (t "f"))
-       (and (marginalia--advised s) "!")))
+       (and (autoloadp (symbol-function s)) "@")
+       (and (marginalia--advised s) "!")
+       (and (get s 'byte-obsolete-info) "-")))
     (when (boundp s)
       (concat
-       (and (get s 'byte-obsolete-variable) "o")
-       (and (local-variable-if-set-p s) "l")
-       (if (custom-variable-p s) "u" "v")
-       (and (ignore-errors (not (equal (symbol-value s) (default-value s)))) "*")))
+       (when (local-variable-if-set-p s)
+         (if (ignore-errors
+               (not (equal (symbol-value s)
+                           (default-value s))))
+             "L" "l"))
+       (if (custom-variable-p s)
+           (if (ignore-errors
+                 (not (equal
+                       (symbol-value s)
+                       (eval (car (get s 'standard-value))))))
+               "U" "u")
+         "v")
+       (and (get s 'byte-obsolete-variable) "-")))
     (and (facep s) "a")
     (and (fboundp 'cl-find-class) (cl-find-class s) "t"))))
 
@@ -407,6 +492,21 @@ t cl-type"
           (substring str (match-end 0))
         str))))
 
+;; Derived from elisp-get-fnsym-args-string
+(defun marginalia--function-args (sym)
+  "Return function arguments for SYM."
+  (let ((tmp))
+    (elisp-function-argstring
+      (cond
+       ((listp (setq tmp (gethash (indirect-function sym)
+                                  advertised-signature-table t)))
+        tmp)
+       ((setq tmp (help-split-fundoc
+		   (ignore-errors (documentation sym t))
+		   sym))
+	(substitute-command-keys (car tmp)))
+       (t (help-function-arglist sym))))))
+
 (defun marginalia-annotate-symbol (cand)
   "Annotate symbol CAND with its documentation string."
   (when-let (sym (intern-soft cand))
@@ -441,16 +541,70 @@ keybinding since CAND includes it."
     ;; Strip until the last whitespace in order to support flat imenu
     (marginalia-annotate-symbol (replace-regexp-in-string "^.* " "" cand))))
 
+(defun marginalia-annotate-function (cand)
+  "Annotate function CAND with its documentation string."
+  (when-let (sym (intern-soft cand))
+    (when (functionp sym)
+      (concat
+       (marginalia-annotate-binding cand)
+       (marginalia--fields
+        ((marginalia--symbol-class sym) :face 'marginalia-type)
+        ((marginalia--function-args sym) :face 'marginalia-value
+         :truncate (/ marginalia-truncate-width 2))
+        ((marginalia--function-doc sym) :truncate marginalia-truncate-width
+         :face 'marginalia-documentation))))))
+
+(defun marginalia--variable-value (sym)
+  "Return the variable value of SYM as string."
+  (cond
+   ((not (boundp sym))
+    (propertize "<unbound>" 'face 'marginalia-null))
+   ((and marginalia-censor-variables
+         (let ((name (symbol-name sym)))
+           (seq-find (lambda (r)
+                       (if (symbolp r)
+                           (eq r sym)
+                         (string-match-p r name)))
+                     marginalia-censor-variables)))
+    (propertize "*****" 'face 'marginalia-null))
+   (t (let ((val (symbol-value sym)))
+        (pcase (symbol-value sym)
+          ('nil (propertize "nil" 'face 'marginalia-null))
+          ('t (propertize "t" 'face 'marginalia-true))
+          ((pred keymapp) (propertize "<keymap>" 'face 'marginalia-value))
+          ((pred hash-table-p) (propertize "<hash-table>" 'face 'marginalia-value))
+          ((and (pred functionp) (pred symbolp))
+           ;; NOTE: We are not consistent here, values are generally printed unquoted. But we
+           ;; make an exception for function symbols to visually distinguish them from symbols.
+           ;; I am not entirely happy with this, but we should not add quotation to every type.
+           (propertize (format "#'%s" val) 'face 'marginalia-function))
+          ((pred symbolp) (propertize (symbol-name val) 'face 'marginalia-symbol))
+          ((pred numberp) (propertize (number-to-string val) 'face 'marginalia-number))
+          (_ (let ((print-escape-newlines t)
+                   (print-escape-control-characters t)
+                   (print-escape-multibyte t)
+                   (print-level 10)
+                   (print-length marginalia-truncate-width))
+               (propertize
+                (prin1-to-string
+                 (if (stringp val)
+                     ;; Get rid of string properties to save some of the precious space
+                     (substring-no-properties
+                      val 0
+                      (min (length val) marginalia-truncate-width))
+                   val))
+                'face
+                (cond
+                 ((listp val) 'marginalia-list)
+                 ((stringp val) 'marginalia-string)
+                 (t 'marginalia-value))))))))))
+
 (defun marginalia-annotate-variable (cand)
   "Annotate variable CAND with its documentation string."
   (when-let (sym (intern-soft cand))
     (marginalia--fields
      ((marginalia--symbol-class sym) :face 'marginalia-type)
-     ((let ((print-escape-newlines t)
-            (print-escape-control-characters t)
-            (print-escape-multibyte t))
-        (prin1-to-string (if (boundp sym) (symbol-value sym) 'unbound)))
-      :truncate (/ marginalia-truncate-width 3) :face 'marginalia-variable)
+     ((marginalia--variable-value sym) :truncate (/ marginalia-truncate-width 2))
      ((documentation-property sym 'variable-documentation)
       :truncate marginalia-truncate-width :face 'marginalia-documentation))))
 
@@ -458,7 +612,7 @@ keybinding since CAND includes it."
   "Annotate environment variable CAND with its current value."
   (when-let (val (getenv cand))
     (marginalia--fields
-     (val :truncate marginalia-truncate-width :face 'marginalia-variable))))
+     (val :truncate marginalia-truncate-width :face 'marginalia-value))))
 
 (defun marginalia-annotate-face (cand)
   "Annotate face CAND with its documentation string and face example."
@@ -544,7 +698,7 @@ keybinding since CAND includes it."
   "Return bookmark type string of BM.
 
 The string is transformed according to `marginalia-bookmark-type-transformers'."
-  (let ((handler (or (alist-get 'handler bm) 'bookmark-default-handler)))
+  (let ((handler (or (bookmark-get-handler bm) 'bookmark-default-handler)))
     ;; Some libraries use lambda handlers instead of symbols. For
     ;; example the function `xwidget-webkit-bookmark-make-record' is
     ;; affected. I consider this bad style since then the lambda is
@@ -559,11 +713,11 @@ The string is transformed according to `marginalia-bookmark-type-transformers'."
 
 (defun marginalia-annotate-bookmark (cand)
   "Annotate bookmark CAND with its file name and front context string."
-  (when-let ((bm (bookmark-get-bookmark-record (assoc cand bookmark-alist))))
-    (let ((front (alist-get 'front-context-string bm)))
+  (when-let ((bm (assoc cand bookmark-alist)))
+    (let ((front (bookmark-get-front-context-string bm)))
       (marginalia--fields
        ((marginalia--bookmark-type bm) :width 10 :face 'marginalia-type)
-       ((alist-get 'filename bm) :truncate 40 :face 'marginalia-file-name)
+       ((bookmark-get-filename bm) :truncate 40 :face 'marginalia-file-name)
        ((if (or (not front) (string= front ""))
             ""
           (concat (string-trim
@@ -588,20 +742,21 @@ The string is transformed according to `marginalia-bookmark-type-transformers'."
   "Annotate coding system CAND with its description."
   (marginalia--documentation (coding-system-doc-string (intern cand))))
 
-(defun marginalia-annotate-buffer (cand)
-  "Annotate buffer CAND with modification status, file name and major mode."
-  (when-let (buffer (get-buffer cand))
-    (marginalia--fields
-     ((format-mode-line '((:propertize "%1*%1+%1@" face marginalia-modified)
-                          marginalia--separator
-                          (7 (:propertize "%I" face marginalia-size))
-                          marginalia--separator
-                          ;; InactiveMinibuffer has 18 letters, but there are longer names.
-                          ;; For example Org-Agenda produces very long mode names.
-                          ;; Therefore we have to truncate.
-                          (20 (-20 (:propertize mode-name face marginalia-mode))))
-                        nil nil buffer))
-     ((if-let (proc (get-buffer-process buffer))
+(defun marginalia--buffer-status (buffer)
+  "Return the status of BUFFER as a string."
+  (format-mode-line '((:propertize "%1*%1+%1@" face marginalia-modified)
+                      marginalia--separator
+                      (7 (:propertize "%I" face marginalia-size))
+                      marginalia--separator
+                      ;; InactiveMinibuffer has 18 letters, but there are longer names.
+                      ;; For example Org-Agenda produces very long mode names.
+                      ;; Therefore we have to truncate.
+                      (20 (-20 (:propertize mode-name face marginalia-mode))))
+                    nil nil buffer))
+
+(defun marginalia--buffer-file (buffer)
+  "Return the file or process name of BUFFER."
+  (if-let (proc (get-buffer-process buffer))
           (format "(%s %s) %s"
                   proc (process-status proc)
                   (abbreviate-file-name (buffer-local-value 'default-directory buffer)))
@@ -615,7 +770,14 @@ The string is transformed according to `marginalia-bookmark-type-transformers'."
                                    (buffer-local-value 'default-directory buffer))))
               ((local-variable-p 'list-buffers-directory buffer)
                (buffer-local-value 'list-buffers-directory buffer)))
-             "")))
+             ""))))
+
+(defun marginalia-annotate-buffer (cand)
+  "Annotate buffer CAND with modification status, file name and major mode."
+  (when-let (buffer (get-buffer cand))
+    (marginalia--fields
+     ((marginalia--buffer-status buffer))
+     ((marginalia--buffer-file buffer)
       :truncate (/ marginalia-truncate-width 2)
       :face 'marginalia-file-name))))
 
@@ -636,29 +798,107 @@ component of a full file path."
       ;; necessary information (there's not much else we can do)
       cand))
 
-(defun marginalia--remote-p (path)
-  "Return t if PATH is a remote path."
-  (string-match-p "\\`/[^/|:]+:" (substitute-in-file-name path)))
+(defun marginalia--remote-protocol (path)
+  "Return the remote protocol of PATH."
+  (save-match-data
+    (setq path (substitute-in-file-name path))
+    (and (string-match "\\`/\\([^/|:]+\\):" path)
+         (match-string 1 path))))
+
+(defun marginalia--annotate-local-file (cand)
+  "Annotate local file CAND."
+  (when-let (attrs (ignore-errors
+                     ;; may throw permission denied errors
+                     (file-attributes (substitute-in-file-name
+                                       (marginalia--full-candidate cand))
+                                      'integer)))
+    (marginalia--fields
+     ((marginalia--file-owner attrs)
+      :width 12 :face 'marginalia-file-owner)
+     ((marginalia--file-modes attrs))
+     ((file-size-human-readable (file-attribute-size attrs))
+      :face 'marginalia-size :width -7)
+     ((marginalia--time (file-attribute-modification-time attrs))
+      :face 'marginalia-date :width -12))))
 
 (defun marginalia-annotate-file (cand)
   "Annotate file CAND with its size, modification time and other attributes.
 These annotations are skipped for remote paths."
-  (if (or (marginalia--remote-p cand)
-          (when-let (win (active-minibuffer-window))
-            (with-current-buffer (window-buffer win)
-              (marginalia--remote-p (minibuffer-contents-no-properties)))))
-      (marginalia--fields ("*Remote*" :face 'marginalia-documentation))
-    (when-let (attributes (file-attributes (substitute-in-file-name (marginalia--full-candidate cand)) 'string))
-      (marginalia--fields
-       ((file-attribute-modes attributes) :face 'marginalia-file-modes)
-       ((format "%s:%s"
-                (file-attribute-user-id attributes)
-                (file-attribute-group-id attributes))
-        :width 12 :face 'marginalia-file-owner)
-       ((file-size-human-readable (file-attribute-size attributes)) :width 7 :face 'marginalia-size)
-       ((format-time-string
-         "%b %d %H:%M"
-         (file-attribute-modification-time attributes)) :face 'marginalia-date)))))
+  (if-let (remote (or (marginalia--remote-protocol cand)
+                      (when-let (win (active-minibuffer-window))
+                        (with-current-buffer (window-buffer win)
+                          (marginalia--remote-protocol (minibuffer-contents-no-properties))))))
+      (marginalia--fields (remote :format "*%s*" :face 'marginalia-documentation))
+    (marginalia--annotate-local-file cand)))
+
+(defun marginalia--file-owner (attrs)
+  "Return file owner given ATTRS."
+  (let ((uid (file-attribute-user-id attrs))
+        (gid (file-attribute-group-id attrs)))
+    (if (or (/= (user-uid) uid) (/= (group-gid) gid))
+        (format "%s:%s" (or (user-login-name uid) uid) (or (group-name gid) gid))
+      "")))
+
+(defun marginalia--file-modes (attrs)
+  "Return fontified file modes given the ATTRS."
+  ;; Without caching this can a be significant portion of the time
+  ;; `marginalia-annotate-file' takes to execute. Caching improves performance
+  ;; by about a factor of 20.
+  (setq attrs (file-attribute-modes attrs))
+  (or (car (member attrs marginalia--fontified-file-modes))
+      (progn
+        (setq attrs (substring attrs)) ;; copy because attrs is about to be modified
+        (dotimes (i (length attrs))
+          (put-text-property
+           i (1+ i) 'face
+           (pcase (aref attrs i)
+             (?- 'marginalia-file-priv-no)
+             (?d 'marginalia-file-priv-dir)
+             (?l 'marginalia-file-priv-link)
+             (?r 'marginalia-file-priv-read)
+             (?w 'marginalia-file-priv-write)
+             (?x 'marginalia-file-priv-exec)
+             ((or ?s ?S ?t ?T) 'marginalia-file-priv-other)
+             (_ 'marginalia-file-priv-rare))
+           attrs))
+        (push attrs marginalia--fontified-file-modes)
+        attrs)))
+
+(defconst marginalia--time-relative
+  `((100 "sec" 1)
+    (,(* 60 100) "min" 60.0)
+    (,(* 3600 30) "hour" 3600.0)
+    (,(* 3600 24 400) "day" ,(* 3600.0 24.0))
+    (nil "year" ,(* 365.25 24 3600)))
+  "Formatting used by the function `marginalia--time-relative'.")
+
+;; Taken from `seconds-to-string'.
+(defun marginalia--time-relative (time)
+  "Format TIME as a relative age."
+  (setq time (float-time (time-since time)))
+  (if (<= time 0)
+      "0 secs ago"
+    (let ((sts marginalia--time-relative) here)
+      (while (and (car (setq here (pop sts))) (<= (car here) time)))
+      (setq time (round time (caddr here)))
+      (format "%s %s%s ago" time (cadr here) (if (= time 1) "" "s")))))
+
+(defun marginalia--time-absolute (time)
+  "Format TIME as an absolute age."
+  (let ((system-time-locale "C"))
+    (format-time-string
+     ;; decoded-time-year is only available on Emacs 27, use nth 5 here.
+     (if (> (nth 5 (decode-time (current-time)))
+            (nth 5 (decode-time time)))
+         " %Y %b %d"
+       "%b %d %H:%M")
+     time)))
+
+(defun marginalia--time (time)
+  "Format file age TIME, suitably for use in annotations."
+  (if (< (float-time (time-since time)) marginalia-max-relative-age)
+      (marginalia--time-relative time)
+    (marginalia--time-absolute time)))
 
 (defmacro marginalia--project-root ()
   "Return project root."
@@ -678,8 +918,8 @@ These annotations are skipped for remote paths."
 
 (defun marginalia-classify-by-command-name ()
   "Lookup category for current command."
-  (and marginalia--this-command
-       (alist-get marginalia--this-command marginalia-command-categories)))
+  (and marginalia--command
+       (alist-get marginalia--command marginalia-command-categories)))
 
 (defun marginalia-classify-original-category ()
   "Return original category reported by completion metadata."
@@ -710,9 +950,11 @@ looking for a regexp that matches the prompt."
   "Setup annotator context with completion METADATA around BODY."
   (declare (indent 1))
   (let ((w (make-symbol "w"))
+        (c (make-symbol "c"))
         (o (make-symbol "o")))
     ;; Take the window width of the current window (minibuffer window!)
     `(let ((marginalia--metadata ,metadata)
+           (,c marginalia--cache)
            (,w (window-width))
            ;; Compute marginalia-align-offset. If the right-fringe-width is
            ;; zero, use an additional offset of 1 by default! See
@@ -724,13 +966,42 @@ looking for a regexp that matches the prompt."
        ;; Otherwise it would probably suffice to only change the current buffer.
        ;; We need the `selected-window' fallback for Embark Occur.
        (with-selected-window (or (minibuffer-selected-window) (selected-window))
-         (let ((marginalia-truncate-width (min (/ ,w 2) marginalia-truncate-width))
-               (marginalia-align-offset (or marginalia-align-offset ,o))
+         (let ((marginalia--cache ,c) ;; Take the cache from the minibuffer
+               (marginalia-truncate-width (min (/ ,w 2) marginalia-truncate-width))
                (marginalia--separator (if (>= ,w marginalia-separator-threshold) "    " " "))
-               (marginalia--margin (when (>= ,w (+ marginalia-margin-min marginalia-margin-threshold))
-                                     (make-string (- ,w marginalia-margin-threshold) 32))))
+               (marginalia--margin
+                (+ (or marginalia-align-offset ,o)
+                   (if (>= ,w (+ marginalia-margin-min marginalia-margin-threshold))
+                       (- ,w marginalia-margin-threshold)
+                     0))))
            ,@body)))))
 
+(defun marginalia--cache-reset ()
+  "Reset the cache."
+  (when marginalia--cache
+    (setq marginalia--cache (and (> marginalia--cache-size 0)
+                                 (cons nil (make-hash-table :test #'equal
+                                                            :size marginalia--cache-size))))))
+
+(defun marginalia--cached (fun key)
+  "Cached application of function FUN with KEY.
+
+The cache keeps around the last `marginalia--cache-size' computed annotations.
+The cache is mainly useful when scrolling in completion UIs like Vertico or
+Selectrum."
+  (if marginalia--cache
+      (let ((ht (cdr marginalia--cache)))
+        (or (gethash key ht)
+            (let ((val (funcall fun key)))
+              (setcar marginalia--cache (cons key (car marginalia--cache)))
+              (puthash key val ht)
+              (when (>= (hash-table-count ht) marginalia--cache-size)
+                (let ((end (last (car marginalia--cache) 2)))
+                  (remhash (cadr end) ht)
+                  (setcdr end nil)))
+              val)))
+    (funcall fun key)))
+
 (defun marginalia--completion-metadata-get (metadata prop)
   "Meant as :before-until advice for `completion-metadata-get'.
 METADATA is the metadata.
@@ -742,7 +1013,7 @@ PROP is the property which is looked up."
                  (annotate (marginalia--annotator cat)))
        (lambda (cand)
          (marginalia--context metadata
-           (funcall annotate cand)))))
+           (marginalia--cached annotate cand)))))
     ('affixation-function
      ;; We do want the advice triggered for `completion-metadata-get'.
      ;; Return wrapper around `annotation-function'.
@@ -750,7 +1021,7 @@ PROP is the property which is looked up."
                  (annotate (marginalia--annotator cat)))
        (lambda (cands)
          (marginalia--context metadata
-           (mapcar (lambda (x) (list x "" (or (funcall annotate x) ""))) cands)))))
+           (mapcar (lambda (x) (list x "" (or (marginalia--cached annotate x) ""))) cands)))))
     ('category
      ;; Find the completion category by trying each of our classifiers.
      ;; Store the metadata for `marginalia-classify-original-category'.
@@ -758,22 +1029,28 @@ PROP is the property which is looked up."
        (run-hook-with-args-until-success 'marginalia-classifiers)))))
 
 (defun marginalia--minibuffer-setup ()
-  "Setup minibuffer for `marginalia-mode'.
+  "Setup the minibuffer for Marginalia.
 Remember `this-command' for `marginalia-classify-by-command-name'."
-  (setq marginalia--this-command this-command))
+  (setq marginalia--cache t marginalia--command this-command)
+  ;; Reset cache if window size changes, recompute alignment
+  (add-hook 'window-state-change-hook #'marginalia--cache-reset nil 'local)
+  (marginalia--cache-reset))
 
 (defun marginalia--base-position (completions)
   "Record the base position of COMPLETIONS."
   ;; NOTE: As a small optimization track the base position only for file completions,
   ;; since `marginalia--full-candidate' is only used for files as of now.
   (when minibuffer-completing-file-name
-    (setq marginalia--base-position (or (cdr (last completions)) 0)))
+    (let ((base (or (cdr (last completions)) 0)))
+      (unless (= marginalia--base-position base)
+        (marginalia--cache-reset)
+        (setq marginalia--base-position base))))
   completions)
 
 ;;;###autoload
 (define-minor-mode marginalia-mode
   "Annotate completion candidates with richer information."
-  :global t
+  :global t :group 'marginalia
   (if marginalia-mode
       (progn
         ;; Ensure that we remember this-command in order to select the annotation function.
@@ -788,7 +1065,7 @@ Remember `this-command' for `marginalia-classify-by-command-name'."
 
 ;;;###autoload
 (defun marginalia-cycle ()
-  "Cycle between annotators in `marginalia-annotators'."
+  "Cycle between annotators in `marginalia-annotator-registry'."
   (interactive)
   (if-let* ((win (active-minibuffer-window))
             (buf (window-buffer win)))
@@ -805,13 +1082,16 @@ Remember `this-command' for `marginalia-classify-by-command-name'."
           (setq cat (assq cat marginalia-annotator-registry))
           (unless cat
             (user-error "Marginalia: No annotators found"))
+          (marginalia--cache-reset)
           (setcdr cat (append (cddr cat) (list (cadr cat))))
           ;; When the builtin annotator is selected and no builtin function is available, skip to
           ;; the next annotator. Note that we cannot use `completion-metadata-get' to access the
           ;; metadata since we must bypass the `marginalia--completion-metadata-get' advice.
           (when (and (eq (cadr cat) 'builtin)
                      (not (assq 'annotation-function metadata))
-                     (not (assq 'affixation-function metadata)))
+                     (not (assq 'affixation-function metadata))
+                     (not (plist-get completion-extra-properties :annotation-function))
+                     (not (plist-get completion-extra-properties :affixation-function)))
             (setcdr cat (append (cddr cat) (list (cadr cat)))))
           (message "Marginalia: Use annotator `%s' for category `%s'" (cadr cat) (car cat))))
     (user-error "Marginalia: No active minibuffer")))
